import secrets
from ..models.session import Session
from flaschengeist.system.database import db
from flaschengeist import logger
from werkzeug.exceptions import Forbidden
from datetime import datetime, timezone
from . import Singleton


class SessionController(metaclass=Singleton):
    """Control all created Sessions

    This Class create, delete, find and manage Sessions.

    Attributes:
        lifetime: Variable for the Lifetime of a Session in seconds.
    """

    def __init__(self, lifetime=1800):
        self.lifetime = lifetime

    def validate_token(self, token, user_agent, permissions):
        """Verify session

        Verify a Session and Roles so if the User has permission or not.
        Retrieves the access token if valid else retrieves False

        Args:
            token: Token to verify.
            user_agent: User agent of browser to check
            permissions: Permissions needed to access restricted routes
        Returns:
            A Session for this given Token or False.
        """
        logger.debug("check token {{ {} }} is valid".format(token))
        access_token = Session.query.filter_by(token=token).one_or_none()
        if access_token:
            logger.debug("token found, check if expired or invalid user agent differs")
            if access_token.expires >= datetime.utcnow() and (
                access_token.browser == user_agent.browser and access_token.platform == user_agent.platform
            ):
                if not permissions or access_token.user.has_permissions(permissions):
                    access_token.refresh()
                    db.session.commit()
                    return access_token
            else:
                logger.debug("access token is out of date or invalid client used")
                self.delete_session(access_token)
        logger.debug("no valid access token with token: {{ {} }} and permissions: {{ {} }}".format(token, permissions))
        return False

    def create(self, user, user_agent=None) -> Session:
        """Create a Session

        Args:
            user: For which User is to create a Session
            user_agent: User agent to identify session

        Returns:
                Session: A created Token for User
        """
        logger.debug("create access token")
        token_str = secrets.token_hex(16)
        session = Session(
            token=token_str, user=user, lifetime=self.lifetime, browser=user_agent.browser, platform=user_agent.platform
        )
        session.refresh()
        db.session.add(session)
        db.session.commit()

        logger.debug("access token is {{ {} }}".format(session.token))
        return session

    def get_session(self, token, owner=None):
        """Retrieves Session from token string

        Args:
            token (str): Token string
            owner (User, optional): User owning the token

        Raises:
            Forbidden: Raised if owner is set but does not match
        Returns:
            Session: Token object identified by given token string
        """
        session = Session.query.filter(Session.token == token).one_or_none()
        if session and (owner and owner != session.user):
            raise Forbidden
        return session

    def get_users_sessions(self, user):
        return Session.query.filter(Session.user == user)

    @staticmethod
    def delete_session(token: Session):
        """Deletes given Session

        Args:
            token (Session): Token to delete
        """
        db.session.delete(token)
        db.session.commit()

    @staticmethod
    def update_session(session):
        session.refresh()
        db.session.commit()

    def set_lifetime(self, session, lifetime):
        session.lifetime = lifetime
        self.update_session(session)

    def clear_expired(self):
        """Remove expired tokens from database"""
        logger.debug("Clear expired Sessions")
        deleted = Session.query.filter(Session.expires < datetime.utcnow()).delete()
        logger.debug("{} sessions have been removed".format(deleted))
        db.session.commit()
